﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography;
using System.Web.Script.Serialization;

namespace StatisticsBoxSDK
{
    [Serializable]
    class AuthData
    {
        public string device_salt;
        public string session_salt;
        public string api_version;
    }

    class AccountData
    {
        public string login;
        public string pass_sha256;
    }

    [Serializable]
    public class SbException : Exception
    {
        public virtual string Id { get; private set; }
        public SbException() { }
        public SbException(string message, string id) : base(message) { 
            this.Id = id;
        }
        public SbException(string message, Exception inner) : base(message, inner) { }
        protected SbException(
          System.Runtime.Serialization.SerializationInfo info,
          System.Runtime.Serialization.StreamingContext context)
            : base(info, context) { }
    }

    public class Device
    {
        public readonly String UniqueDeviceName;
        public readonly String FriendlyName;
        public readonly String PresentationURL;
        public readonly String ModelName;
        public readonly String ModelNumber;
        public readonly String SerialNumber;

        private WebClientEx _client;
        private JavaScriptSerializer _serializer = new JavaScriptSerializer();
        private UrlCreator _urlCreator = new UrlCreator("");
        private bool _isConnected;
        private String _lastUserName;
        private String _lastPassword;

        public static Device CreateFromHost(String hostNameOrIp)
        {
            //To get full device info download UPNP xml info from fixed port on the host
            //Now we haven't fixed port with UPNP xml info, so leave all info empty
            return new Device("", "", "http://" + hostNameOrIp + "/", "", "", "");
        }

        public Device(String uniqueDeviceName, String friendlyName, String presentationURL,
               String modelName, String modelNumber, String serialNumber)
        {
            UniqueDeviceName = uniqueDeviceName;
            FriendlyName = friendlyName;
            PresentationURL = presentationURL;
            ModelName = modelName;
            ModelNumber = modelNumber;
            SerialNumber = serialNumber;
            _isConnected = false;
            _serializer.RegisterConverters(new JavaScriptConverter[] { 
                    new AlgoParametersConverter(),
                    new DeviceSettingsConverter() });
        }

        public void Connect(String UserName, String Password)
        {
            ServicePointManager.ServerCertificateValidationCallback =
                new RemoteCertificateValidationCallback(
                    delegate
                    { return true; }
                );
            _client = new WebClientEx();
            _client.Proxy = null;
            _client.BaseAddress = PresentationURL;
            _lastUserName = UserName;
            _lastPassword = Password;
            var auth_data = DeserializeFromUrl<AuthData>("new_session.html");
            if (auth_data.api_version == null || !auth_data.api_version.Equals("3.1"))
                throw new Exception("Unknown authentication method");
            string secret = Utf8Sha256(Password + UserName + auth_data.device_salt);
            AccountData accountData = new AccountData();
            accountData.login = UserName;
            accountData.pass_sha256 = Utf8Sha256(secret + auth_data.session_salt);
            SerializeToUrl("scripts/login.php", accountData);
        }

        private void CheckResponse(byte[] response)
        {
            String responseString = Encoding.UTF8.GetString(response);
            Dictionary<String, Object> json_object = _serializer.DeserializeObject(responseString) as Dictionary<String, Object>;
            if (json_object != null)
            {
                Object error_message, error_id;
                json_object.TryGetValue("error", out error_message);
                json_object.TryGetValue("error_id", out error_id);
                if (error_id == null)
                {
                    throw new Exception("Incorrect server response");
                }
                if (!error_id.Equals(""))
                {
                    throw new SbException(error_message as String, error_id as String);
                }
            }
        }

        private TRet DeserializeFromUrl<TRet>(string url)
        {
            byte[] web_result = _client.DownloadData(url);
            CheckResponse(web_result);
            var result = System.Text.Encoding.UTF8.GetString(web_result);
            return _serializer.Deserialize<TRet>(result);
        }

        private void SerializeToUrl<TPar>(string url, TPar obj)
        {
            url = _urlCreator.ObjectToUrl(url, obj);
            byte[] web_result = _client.DownloadData(url);
            CheckResponse(web_result);
        }

        public HostnameSettings GetHostname()
        {
            return DeserializeFromUrl<HostnameSettings>("scripts/hostname.php");
        }

        public void SetHostname(HostnameSettings hostnameSettings)
        {
            SerializeToUrl("scripts/hostname.php", hostnameSettings);
        }

        public WebserverSettings GetWebserverSettings()
        {
            return DeserializeFromUrl<WebserverSettings>("scripts/webserver.php");
        }

        public void SetWebserverSettings(WebserverSettings webserverSettings)
        {
            SerializeToUrl("scripts/webserver.php", webserverSettings);
        }

        public AlgoParameters GetAlgoParams()
        {
            return DeserializeFromUrl<AlgoParameters>("scripts/algo_params.php");
        }

        public void SetAlgoParams(AlgoParameters algoParams)
        {
            SerializeToUrl("scripts/algo_params.php", algoParams);
        }

        public DeviceSettings GetDeviceSettings()
        {
            return DeserializeFromUrl<DeviceSettings>("scripts/device.php");
        }

        public void SetDeviceSettings(DeviceSettings dev_settings)
        {
            SerializeToUrl("scripts/device.php", dev_settings);
        }

        public NetSettings GetNetSettings()
        {
            return DeserializeFromUrl<NetSettings>("scripts/net.php");
        }

        public void SetNetSettings(NetSettings net_settings)
        {
            SerializeToUrl("scripts/net.php", net_settings);
        }

        public void UploadUpdate(String file_name)
        {
            byte[] web_result = _client.UploadFile("scripts/upload_update.php", file_name);
            CheckResponse(web_result);
        }

        class ReportRequest
        {
            public string export_type;
            public string how;
            public string from;
            public string time_from;
            public string to;
            public string time_to;
        }

        public string GetReport(String type, DateTime from, DateTime to)
        {
            ReportRequest reportRequest = new ReportRequest();
            reportRequest.export_type = type;
            reportRequest.how = "bytime";
            reportRequest.from = from.ToString("yyyy-MM-dd");
            reportRequest.time_from = from.ToString("HH:mm:ss");
            reportRequest.to = to.ToString("yyyy-MM-dd");
            reportRequest.time_to = to.ToString("HH:mm:ss");
            string url = _urlCreator.ObjectToUrl("scripts/xmlGen.php", reportRequest);
            byte[] web_result = _client.DownloadData(url);
            return System.Text.Encoding.UTF8.GetString(web_result);
        }

        class ClearRequest
        {
            public string how;
            public string from;
            public string time_from;
            public string to;
            public string time_to;
        }

        public void ClearDb(DateTime from, DateTime to)
        {
            ClearRequest clearRequest = new ClearRequest();
            clearRequest.how = "bytime";
            clearRequest.from = from.ToString("yyyy-MM-dd");
            clearRequest.time_from = from.ToString("HH:mm:ss");
            clearRequest.to = to.ToString("yyyy-MM-dd");
            clearRequest.time_to = to.ToString("HH:mm:ss");
            SerializeToUrl("scripts/cleardb.php", clearRequest);
        }

        public void Disconnect()
        {
            _client.Dispose();
        }
        public bool IsConnected()
        {
            return _isConnected;
        }

        private String Utf8Sha256(String text)
        {
            byte[] binarray = (new SHA256CryptoServiceProvider()).ComputeHash(Encoding.UTF8.GetBytes(text));
            String result = "";
            for (int i = 0; i < binarray.Length; i++)
            {
                result += binarray[i].ToString("x2");
            }
            return result;
        }
    }
}
